import { isNothing, isNull, isUndefined } from './is';
import { curry2, curry2n, curry3, curry4n } from './curry';
// Composition
export const compose = (...fns) => (...args) => fns.reduceRight((x, f, i) => i === fns.length - 1 ? f(...args) : f(x), null);
export const pipe = (...fns) => (...args) => (Array.isArray(fns[0]) ? fns[0] : fns).reduce((x, f, i) => i === 0 ? f(...x) : f(x), args);
export const converge = (f, ...fns) => (x) => f(...fns.map(g => g(x)));
export const spread = (f, ...fns) => (...x) => f(...fns.map((g, i) => g(x[i])));
export const spreadLeft = (f, ...fns) => (...x) => f(...fns.map((g, i) => g(x[fns.length - i - 1])));
export const applyUsing = curry2((f, x) => f(x)(x));
export const flip = curry3((f, a, b) => f(b, a));
// Logic
export const ifElse = curry4n((f, g, h, ...x) => f(...x) ? g(...x) : h(...x));
export const negate = curry2((f, x) => !f(x));
export const equals = curry2((a, b) => a === b);
export const or = curry3((f, g, x) => f(x) || g(x));
export const and = curry3((f, g, x) => f(x) && g(x));
// tslint:disable-next-line:variable-name
export const any = (...fns) => (x) => fns.some(f => f(x));
export const mapIf = curry3((p, f, x) => p(x) ? f(x) : x);
export const mapUnless = curry3((p, f, x) => p(x) ? x : f(x));
// Transformation
export const map = curry2((f, x) => f(x));
export const defaultUsing = curry2((f, x) => isUndefined(x) ? f() : x);
export const defaultIfUndefined = curry2((a, x) => isUndefined(x) ? a : x);
export const defaultIfNothing = curry2((a, x) => isNothing(x) ? a : x);
export const defaultIfNull = curry2((a, x) => isNull(x) ? a : x);
export const defaultIf = curry3((f, a, x) => f(x) ? a : x);
export const defaultUnless = curry3((f, a, x) => f(x) ? x : a);
export const valueOrDefault = curry2((x, a) => isUndefined(x) ? a : x);
export const undefinedIfNull = defaultIfNull(void 0);
export const identity = (x) => x;
export const noop = (...args) => { };
export const constant = (x) => () => x;
export const nothing = constant(void 0);
export const tap = (f) => (x) => (f(x), x);
// Errors
export const throwError = (s = '') => { throw new Error(s); };
export const throwNotImplemented = (s) => throwError(`Not implemented${s ? `: ${s}` : ''}`);
export const throwNotSupported = (s) => throwError(`Operation not supported${s ? `: ${s}` : ''}`);
export const throwInvalidOperation = (s) => throwError(`Invalid operation${s ? `: ${s}` : ''}`);
export const throwMissing = (name, message) => throwError(`No value is defined${name ? ` for "${name}"${message ? ` (${message})` : ''}` : ''}`);
export const throwArgumentError = (name, message) => throwError(`Invalid ${name ? `value for parameter "${name}"` : 'argument value'}${message ? `: ${message}` : ''}`);
export const error = (s) => () => throwError(s);
export const notImplemented = (s) => () => throwNotImplemented(s);
export const notSupported = (s) => () => throwNotSupported(s);
export const invalidOperation = (s) => () => throwInvalidOperation(s);
export const missing = (name, message) => () => throwMissing(name, message);
export const argumentError = (name, message) => () => throwArgumentError(name, message);
// Function Arguments
export const getArg = (i) => function () { return arguments[i]; };
export const argsToArray = (...args) => args;
export const mapApply = (f, g) => (...args) => f(...args.map(g));
// Function Binding
export const invoker = (f, ...a) => (...b) => f(...a, ...b);
export const invokeWithArgs = (...args) => (f) => f(...args);
export const invokeMethod = (key) => (target) => target[key]();
export const invokeMethodWithArgs = curry2n((key, ...args) => (target) => target[key](...args));
// Object
export const assoc = curry3((key, value, target) => (Object.assign({}, target, { [key]: value })));
export const objectOf = curry2((key, value) => ({ [key]: value }));
export const construct = curry2n((type, ...args) => new type(...args));
export const mapProps = curry2((f, o) => Object.keys(o).map(key => f(key, o[key])));
export const mapObject = curry2((f, o) => Object.keys(o).reduce((o, key) => (o[key] = f(o[key], key), o), Object.assign({}, o)));
export const toPairs = (o) => Object.keys(o).map(key => [key, o[key]]);
export const get = curry2((key, o) => o[key]);
export const has = curry2((key, o) => key in o);
export const maybeGet = curry2((key, o) => isNothing(o) ? void 0 : o[key]);
export const getFrom = curry2((o, k) => o[k]);
export const valueOf = (value) => isNothing(value) ? value : value.valueOf();
// String
export const substr = (from, length) => (s) => s.substr(from, length);
//# sourceMappingURL=core.js.map